/*
 * Copyright (C) 1997-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file server.c Paraslash's main server */


/** \mainpage Paraslash API Reference
 *
 *  Good starting points for reading are probably \ref audio_file_selector,
 *  \ref sender, \ref receiver, \ref receiver_node, \ref filter, \ref
 *  filter_node.
 *
 */



#include "server.cmdline.h"
#include "db.h"
#include "server.h"
#include "afs.h"
#include "config.h"
#include "close_on_fork.h"
#include "send.h"
#include "error.h"
#include "net.h"
#include "daemon.h"
#include "string.h"
#include "ipc.h"

/** define the array of error lists needed by para_server */
INIT_SERVER_ERRLISTS;

/** shut down non-authorized connections after that many seconds */
#define ALARM_TIMEOUT 10

/* these are exported to afs.c. command.c and to all selectors */
struct misc_meta_data *mmd;
/** the configuration of para_server
 *
 * It also contains the options for all audio file selectors and all supported
 * senders.
*/
struct gengetopt_args_info conf;
char *user_list = NULL;
extern void http_send_init(struct sender *);
extern void ortp_send_init(struct sender *);
extern struct audio_format afl[];

/** the list of supported audio file selectors */
struct audio_file_selector dblist[] = {
	{
		.name = "random",
		.init = random_selector_init,
		.update_audio_file = NULL,
	},
	{
		.name = "playlist",
		.init = playlist_selector_init,
		.update_audio_file = NULL,
		.pre_select = NULL,
		.post_select = NULL,
	},
#ifdef HAVE_MYSQL
	{
		.name = "mysql",
		.init = mysql_selector_init,
		.update_audio_file = NULL,
		.pre_select = NULL,
		.post_select = NULL,
	},
#endif
	{
		.name = NULL,
	}
};

/** the list of supported senders */
struct sender senders[] = {
	{
		.name = "http",
		.init = http_send_init,
	},
#ifdef HAVE_ORTP
	{
		.name = "ortp",
		.init = ortp_send_init,
	},
#endif
	{
		.name = NULL,
	}
};


/* global variables for server-internal use */
static FILE *logfile;
static int mmd_mutex, mmd_shm_id;
static int signal_pipe;

/**
 * para_server's log function
 *
 * \param ll the log level
 * \param fmt the format string describing the log message
 */
void para_log(int ll, char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[MAXLINE] = "";
	pid_t mypid;

	if (ll < conf.loglevel_arg)
		return;
	if (!logfile) {
		if (ll < WARNING)
			outfd = stdout;
		else
			outfd = stderr;
	} else
		outfd = logfile;
	if (conf.daemon_given && !logfile)
		return;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, MAXLINE, "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s ", str);
	if (conf.loglevel_arg <= INFO)
		fprintf(outfd, "%i: ", ll);
	mypid = getpid();
	if (conf.loglevel_arg <= INFO)
		fprintf(outfd, "(%d) ", mypid);
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}

/*
 * setup shared memory area and get mutex for locking
 */
static void shm_init(void)
{
	void *shm;
	int ret = shm_new(sizeof(struct misc_meta_data));

	if (ret < 0)
		goto err_out;

	ret = shm_attach(ret, ATTACH_RW, &shm);
	if (ret < 0)
		goto err_out;
	mmd = shm;
	mmd_shm_id = ret;

	ret = mutex_new();
	if (ret < 0)
		goto err_out;
	mmd_mutex = ret;

	mmd->selector_num = 0;
	mmd->num_played = 0;
	mmd->num_commands = 0;
	mmd->events = 0;
	mmd->num_connects = 0;
	mmd->active_connections = 0;
	strcpy(mmd->filename, "(none)");
	mmd->audio_format = -1;
	mmd->afs_status_flags = AFS_NEXT;
	mmd->new_afs_status_flags = AFS_NEXT;
	mmd->sender_cmd_data.cmd_num = -1;
	return;
err_out:
	PARA_EMERG_LOG("%s", PARA_STRERROR(-ret));
	exit(EXIT_FAILURE);
}

/**
 * lock the shared memory area containing the mmd struct
 *
 * \sa semop(2), struct misc_meta_data
 */
void mmd_lock(void)
{
	mutex_lock(mmd_mutex);
}

/**
 * unlock the shared memory area containing the mmd struct
 *
 * \sa semop(2), struct misc_meta_data
 */

void mmd_unlock(void)
{
	mutex_unlock(mmd_mutex);
}

static void parse_config(int override)
{
	char *home = para_homedir();
	struct stat statbuf;
	int ret;
	char *cf;

	if (conf.config_file_given)
		cf = conf.config_file_arg;
	else
		cf = make_message("%s/.paraslash/server.conf", home);
	free(user_list);
	if (!conf.user_list_given)
		user_list = make_message("%s/.paraslash/server.users", home);
	else
		user_list = para_strdup(conf.user_list_arg);
	ret = stat(cf, &statbuf);
	if (ret && conf.config_file_given) {
		ret = -1;
		PARA_EMERG_LOG("can not stat config file %s\n", cf);
		goto out;
	}
	if (!ret) {
		int tmp = conf.daemon_given;
		cmdline_parser_configfile(cf, &conf, override, 0, 0);
		conf.daemon_given = tmp;
	}
	/* logfile */
	if (!conf.logfile_given && conf.daemon_given) {
		ret = -1;
		PARA_EMERG_LOG("%s", "daemon, but no log file\n");
		goto out;
	}
	if (conf.logfile_given)
		logfile = open_log(conf.logfile_arg);
	ret = 1;
out:
	free(cf);
	free(home);
	if (ret > 0)
		return;
	free(user_list);
	user_list = NULL;
	exit(EXIT_FAILURE);
}

static void setup_signal_handling(void)
{
	int ret = 0;

	signal_pipe = para_signal_init();
//	fcntl(signal_pipe, F_SETFL, O_NONBLOCK);
	PARA_NOTICE_LOG("%s", "setting up signal handlers\n");
	ret += para_install_sighandler(SIGINT);
	ret += para_install_sighandler(SIGTERM);
	ret += para_install_sighandler(SIGHUP);
	ret += para_install_sighandler(SIGCHLD);
	ret += para_install_sighandler(SIGUSR1);
	signal(SIGPIPE, SIG_IGN);
	if (ret != 5) {
		PARA_EMERG_LOG("%s", "could not install signal handlers\n");
		exit(EXIT_FAILURE);
	}
}

static void init_selector(void)
{
	int i, ret;

	mmd->selector_change = -1; /* no change nec., set to new num by com_chs */
	if (!conf.selector_given)
		goto random;
	for (i = 0; dblist[i].name; i++) {
		if (strcmp(dblist[i].name, conf.selector_arg))
			continue;
		PARA_NOTICE_LOG("initializing %s audio file selector\n",
			dblist[i].name);
		ret = dblist[i].init(&dblist[i]);
		if (ret < 0) {
			PARA_WARNING_LOG("%s", PARA_STRERROR(-ret));
			break;
		}
		mmd->selector_num = i;
		return;
	}
	PARA_WARNING_LOG("%s", "falling back to the random selector\n");
random:
	mmd->selector_num = 0;
	dblist[0].init(&dblist[0]); /* always successful */
}

static unsigned init_network(void)
{
	int sockfd = init_tcp_socket(conf.port_arg);

	if (sockfd < 0)
		exit(EXIT_FAILURE);
	return sockfd;
}

static void init_random_seed(void)
{
	int fd, ret = -1, len = sizeof(unsigned int);
	unsigned int seed;

	fd = open("/dev/random", O_RDONLY);
	if (fd < 0)
		goto out;
	ret = -2;
	if (read(fd, &seed, len) != len)
		goto out;
	srandom(seed);
	ret = 1;
out:
	if (fd >= 0)
		close(fd);
	if (ret > 0)
		return;
	PARA_EMERG_LOG("can not seed pseudo random generator (ret = %d)\n",
		ret);
	exit(EXIT_FAILURE);
}

static unsigned do_inits(int argc, char **argv)
{
	/* connector's address information */
	int sockfd;

	init_random_seed();
	/* parse command line options */
	cmdline_parser(argc, argv, &conf);
	para_drop_privileges(conf.user_arg);
	/* parse config file, open log and set defaults */
	parse_config(0);
	log_welcome("para_server", conf.loglevel_arg);
	shm_init(); /* init mmd struct */
	server_uptime(UPTIME_SET); /* reset server uptime */
	/* become daemon */
	if (conf.daemon_given)
		daemon_init();
	init_selector();
	PARA_NOTICE_LOG("%s", "initializing audio file sender\n");
	/* audio file sender */
	afs_init();
	mmd->server_pid = getpid();
	setup_signal_handling();
	mmd_lock();
	/* init network socket */
	PARA_NOTICE_LOG("%s", "initializing tcp command socket\n");
	sockfd = init_network();
	if (conf.autoplay_given) {
		mmd->afs_status_flags |= AFS_PLAYING;
		mmd->new_afs_status_flags |= AFS_PLAYING;
	}
	PARA_NOTICE_LOG("%s", "init complete\n");
	return sockfd;
}

static void change_selector(void)
{
	int ret, old = mmd->selector_num, new = mmd->selector_change;

	dblist[old].shutdown();
	ret = dblist[new].init(&dblist[new]);
	mmd->selector_change = -1; /* reset */
	if (ret >= 0) {
		mmd->selector_num = new;
		return;
	}
	/* init failed */
	PARA_ERROR_LOG("%s -- switching to the random selector\n", PARA_STRERROR(-ret));
	dblist[0].init(&dblist[0]);
	mmd->selector_num = 0;
}

/*
 * called when server gets SIGHUP or when client invokes hup command.
 */
static void handle_sighup(void)
{
	PARA_NOTICE_LOG("%s", "SIGHUP\n");
	close_log(logfile); /* gets reopened if necessary by parse_config */
	logfile = NULL;
	parse_config(1); /* reopens log */
	mmd->selector_change = mmd->selector_num; /* do not change selector.. */
	change_selector(); /* .. just reload */
}

static void status_refresh(void)
{
	static int prev_uptime = -1, prev_events = -1;
	int uptime = server_uptime(UPTIME_GET), ret = 1;

	if (prev_events != mmd->events)
		goto out;
	if (mmd->new_afs_status_flags != mmd->afs_status_flags)
		goto out;
	if (uptime / 60 != prev_uptime / 60)
		goto out;
	ret = 0;
out:
	prev_uptime = uptime;
	prev_events = mmd->events;
	mmd->afs_status_flags = mmd->new_afs_status_flags;
	if (ret) {
		PARA_DEBUG_LOG("%d events, forcing status update, af = %d\n",
			mmd->events, mmd->audio_format);
		killpg(0, SIGUSR1);
	}
}

/*
 * MAIN
 */
int main(int argc, char *argv[])
{
	/* listen on sock_fd, new connection on new_fd */
	int sockfd, new_fd;
	struct sockaddr_in their_addr;
	int err, i, max_fileno, ret;
	pid_t chld_pid;
	fd_set rfds, wfds;
	struct timeval *timeout;

	valid_fd_012();
	sockfd = do_inits(argc, argv);
repeat:
	/* check socket and signal pipe in any case */
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	FD_SET(sockfd, &rfds);
	max_fileno = sockfd;
	FD_SET(signal_pipe, &rfds);
	max_fileno = MAX(max_fileno, signal_pipe);

	timeout = afs_preselect();
	status_refresh();
	for (i = 0; senders[i].name; i++) {
		if (senders[i].status != SENDER_ON)
			continue;
		if (!senders[i].pre_select)
			continue;
		senders[i].pre_select(mmd->audio_format >= 0?
			&afl[mmd->audio_format] : NULL,
			&max_fileno,
			&rfds, &wfds);
	}
	if (dblist[mmd->selector_num].pre_select) {
		ret = dblist[mmd->selector_num].pre_select(&rfds, &wfds);
		max_fileno = MAX(max_fileno, ret);
	}
	mmd_unlock();
//	PARA_DEBUG_LOG("%s: select (max = %i)\n", __func__, max_fileno);
	ret = select(max_fileno + 1, &rfds, &wfds, NULL, timeout);
	err = errno;
	//PARA_DEBUG_LOG("%s: select returned  %i\n", __func__, ret);
	mmd_lock();
	if (mmd->selector_change >= 0)
		change_selector();
	if (dblist[mmd->selector_num].post_select)
		dblist[mmd->selector_num].post_select(&rfds, &wfds);
	if (ret < 0 && err == EINTR)
		goto repeat;
	if (ret < 0) {
		PARA_CRIT_LOG("select error (%s)\n", strerror(err));
		goto repeat;
	}
	for (i = 0; senders[i].name; i++) {
		if (senders[i].status != SENDER_ON)
			continue;
		if (!senders[i].post_select)
			continue;
		senders[i].post_select(mmd->audio_format >= 0?
			&afl[mmd->audio_format] : NULL,
			&rfds, &wfds);
	}
	if (!ret) {
		afs_send_chunk();
		status_refresh();
	}
	if (FD_ISSET(signal_pipe, &rfds)) {
		int sig;
		sig = para_next_signal();
		switch (sig) {
		case SIGHUP:
			handle_sighup();
			break;
		case SIGCHLD:
			para_reap_children();
			break;
		/* die on sigint/sigterm. Kill all children too. */
		case SIGINT:
		case SIGTERM:
			PARA_EMERG_LOG("terminating on signal %d\n", sig);
			kill(0, SIGTERM);
			dblist[mmd->selector_num].shutdown();
			mutex_destroy(mmd_mutex);
			shm_detach(mmd);
			shm_destroy(mmd_shm_id);
			exit(EXIT_FAILURE);
		}
	}
	if (mmd->sender_cmd_data.cmd_num >= 0) {
		int num = mmd->sender_cmd_data.cmd_num,
			s = mmd->sender_cmd_data.sender_num;

		if (senders[s].client_cmds[num])
			senders[s].client_cmds[num](&mmd->sender_cmd_data);
		mmd->sender_cmd_data.cmd_num = -1;
	}
	if (!FD_ISSET(sockfd, &rfds))
		goto repeat;

	new_fd = para_accept(sockfd, &their_addr, sizeof(struct sockaddr_in));
	if (new_fd < 0)
		goto repeat;
	PARA_INFO_LOG("got connection from %s, forking\n",
		inet_ntoa(their_addr.sin_addr));
	mmd->num_connects++;
	mmd->active_connections++;
	random();
	chld_pid = fork();
	if (chld_pid < 0) {
		PARA_CRIT_LOG("%s", "fork failed\n");
		goto repeat;
	}
	if (chld_pid) {
		close(new_fd);
		/* parent keeps accepting connections */
		goto repeat;
	}
	alarm(ALARM_TIMEOUT);
	close_listed_fds();
	close(sockfd); /* child doesn't need the listener */
	/*
	 * put info on who we are serving into argv[0] to make
	 * client ip visible in top/ps
	 */
	for (i = argc - 1; i >= 0; i--)
		memset(argv[i], 0, strlen(argv[i]));
	sprintf(argv[0], "para_server (serving %s)",
		inet_ntoa(their_addr.sin_addr));
	return handle_connect(new_fd, &their_addr);
}
